# -*- coding: utf-8 -*-
"""
Core MOO functionality, object model, verbs.
"""

import logging
import warnings
from typing import Union

from .code import context
from .exceptions import QuotaError, AmbiguousObjectError, UserError

__all__ = ["lookup", "create", "write", "invoke", "set_task_perms", "api"]

log = logging.getLogger(__name__)


def lookup(x: Union[int, str]):
    """
    Lookup an object globally by PK, name, or alias.

    :param x: lookup value
    :return: the result of the lookup
    :rtype: Object
    :raises Object.DoesNotExist: when a result cannot be found
    """
    from .models import Object

    if isinstance(x, int):
        return Object.objects.get(pk=x)
    elif isinstance(x, str):
        qs = Object.objects.filter(name__iexact=x)
        aliases = Object.objects.filter(aliases__alias__iexact=x)
        qs = qs.union(aliases)
        if not qs:
            raise Object.DoesNotExist(x)
        return qs[0]
    else:
        raise ValueError(f"{x} is not a supported lookup value.")


def create(name, *a, **kw):
    """
    Creates and returns a new object whose parents are `parents` and whose owner is as described below.
    Provided `parents` are valid Objects with `derive` permission, otherwise :class:`.PermissionError` is
    raised. After the new object is created, its `initialize` verb, if any, is called with no arguments.

    The owner of the new object is either the programmer (if `owner` is not provided), or the provided owner,
    if the caller has permission to `entrust` the object.

    If the intended owner of the new object has a property named `ownership_quota` and the value of that
    property is an integer, then `create()` treats that value as a quota. If the quota is less than
    or equal to zero, then the quota is considered to be exhausted and `create()` raises :class:`.QuotaError` instead
    of creating an object. Otherwise, the quota is decremented and stored back into the `ownership_quota`
    property as a part of the creation of the new object.

    :param name: canonical name
    :type name: str
    :param owner: owner of the Object being created
    :type owner: Object
    :param location: where to create the Object
    :type location: Object
    :param parents: a list of parents for the Object
    :type parents: list[Object]
    :return: the new object
    :rtype: Object
    :raises PermissionError: if the caller is not allowed to `derive` from the parent
    :raises QuotaError: if the caller has a quota and it has been exceeded
    """
    from .models.object import Object, Property
    system = Object.objects.get(pk=1)
    default_parents = [system.root_class] if system.has_property("root_class") else []
    if api.caller:
        try:
            quota = api.caller.get_property("ownership_quota", recurse=False)
            if quota > 0:
                api.caller.set_property("ownership_quota", quota - 1)
            else:
                raise QuotaError(f"{api.caller} has run out of quota.")
        except Property.DoesNotExist:
            pass
        if "owner" not in kw:
            kw["owner"] = api.caller
    if "location" not in kw and "owner" in kw:
        kw["location"] = kw["owner"].location
    if " to " in name:
        print('its a door')
    elif name == "Another Room":
        print('its another room')
    parents = kw.pop("parents", default_parents)
    obj = Object.objects.create(name=name, *a, **kw)
    if parents:
        obj.parents.add(*parents)
    if obj.has_verb("initialize"):
        invoke(obj.get_verb("initialize"))
    return obj


def write(obj, message):
    """
    Send an asynchronous message to the user.

    :param obj: the Object to write to
    :type obj: Object
    :param message: any pickle-able object
    :type message: Any
    """
    from redis.exceptions import ConnectionError  # pylint: disable=redefined-builtin
    from .models.auth import Player

    player = Player.objects.get(avatar=obj)
    from kombu import Exchange, Queue

    from ..celery import app

    with app.default_connection() as conn:
        try:
            channel = conn.channel()
        except ConnectionError:
            warnings.warn(RuntimeWarning(f"ConnectionError({obj}): {message}"))
            return
        queue = Queue(
            "messages", Exchange("moo", type="direct", channel=channel), f"user-{player.user.pk}", channel=channel
        )
        with app.producer_or_acquire() as producer:
            producer.publish(
                dict(message=message, caller=context.get("caller")),
                serializer="pickle",
                exchange=queue.exchange,
                routing_key=f"user-{player.user.pk}",
                declare=[queue],
                retry=True,
            )


def invoke(*args, verb=None, callback=None, delay: int = 0, periodic: bool = False, cron: str = None, **kwargs):
    """
    Asynchronously execute a Verb, optionally returning the result to another Verb.
    This is often a better alternative than using `__call__`-syntax to invoke
    a verb directly, since Verbs invoked this way will each have their own timeout.

    :param verb: the Verb to execute
    :type verb: Verb
    :param callback: an optional callback Verb to receive the result
    :type callback: Verb
    :param delay: seconds to wait before executing, cannot be used with `cron`
    :param periodic: should this task continue to repeat? cannot be used with `cron`
    :param cron: a crontab expression to schedule Verb execution
    :param args: positional arguments for the Verb, if any
    :param kwargs: keyword arguments for the Verb, if any
    :returns: a :class:`.PeriodicTask` instance or `None` if the task is a one-shot
    :rtype: Optional[:class:`.PeriodicTask`]
    """
    from django_celery_beat.models import CrontabSchedule, IntervalSchedule, PeriodicTask

    from moo.core import tasks
    # if hasattr(verb, "invoked_name"):
    #     l = list(args)
    #     l.insert(0, verb.invoked_name)
    #     args = tuple(l)
    if hasattr(verb, "invoked_object"):
        kwargs["this_id"] = verb.invoked_object.pk
    kwargs.update(
        dict(
            caller_id=api.caller.pk,
            verb_id=verb.pk,
            callback_verb_id=callback.pk if callback else None,
        )
    )
    if delay and periodic:
        schedule, _ = IntervalSchedule.objects.get_or_create(
            every=delay,
            period=IntervalSchedule.SECONDS,
        )
        return PeriodicTask.objects.create(
            interval=schedule,
            description=f"{api.caller.pk}:{verb}",
            task="moo.core.tasks.invoke_verb",
            args=args,
            kwargs=kwargs,
        )
    elif cron:
        cronparts = cron.split()
        schedule, _ = CrontabSchedule.objects.get_or_create(
            minute=cronparts[0],
            hour=cronparts[1],
            day_of_week=cronparts[2],
            day_of_month=cronparts[3],
            month_of_year=cronparts[4],
        )
        return PeriodicTask.objects.create(
            interval=schedule,
            description=f"{api.caller.pk}:{verb}",
            task="moo.core.tasks.invoke_verb",
            args=args,
            kwargs=kwargs,
        )
    else:
        tasks.invoke_verb.apply_async(args, kwargs, countdown=delay)
        return None

def set_task_perms(who):
    """
    Set the task permissions to those of `who`.
    :param who: the Object whose permissions to assume
    :type who: Object
    """
    context.override_caller(who)

class _API:
    """
    This wrapper class makes it easy to use a number of contextvars.
    """

    class descriptor:
        """
        Used to perform dynamic lookups of contextvars.
        """

        def __init__(self, name):
            self.name = name

        def __get__(self, obj, objtype=None):
            return context.get(self.name)

    caller = descriptor("caller")  # Code runs with the permission of this object
    player = descriptor("player")  # This object that originally invoked this session, defaults to original caller
    writer = descriptor("writer")  # A callable that will print to the player's console
    parser = descriptor("parser")
    task_id = descriptor("task_id")  # The current task ID


api = _API()



import logging
from time import time

from moo.core import bootstrap, code, create, lookup
from moo.core.models import Player, User

log = logging.getLogger(__name__)

repo = bootstrap.initialize_dataset("default")
wizard = lookup("Wizard")
user = User.objects.create(username="wizard")
Player.objects.create(user=user, avatar=wizard, wizard=True)

containers = lookup("container class")
containers.name = "Generic Container"
containers.save()

with code.context(wizard, log.info):
    sys = lookup(1)
    sys.set_property("player_start", None)
    sys.set_property("container", containers)

    root = create("Root Class", location=None)
    sys.set_property("root_class", root)
    root.add_verb("accept", code="return True")
    root.set_property("description", "The root class from which all objects inherit.", inherited=True)
    root.set_property("key", None, inherited=True)

    rooms = create("Generic Room", parents=[root], location=None)
    sys.set_property("room", rooms)
    rooms.set_property("blessed_object", None, inherited=True)
    rooms.set_property("blessed_task_id", None, inherited=True)
    rooms.set_property("description", "There's not much to see here.", inherited=True)
    rooms.set_property("exits", [], inherited=True)
    rooms.set_property("victim_ejection_message", "You have been ejected from %s by %s.", inherited=True)
    rooms.set_property("ejection_message", "You have ejected %s.", inherited=True)
    rooms.set_property("room_ejection_message", "%s has been ejected by %s.", inherited=True)
    rooms.set_property("residents", [], inherited=True)
    rooms.set_property("free_entry", True, inherited=True)
    rooms.set_property("entrances", [], inherited=True)
    rooms.set_property("dark", False, inherited=True)
    rooms.set_property("content_list_type", 0, inherited=True)

    mail_room = create("Mail Distribution Center", parents=[rooms], location=None)

    player = create("Generic Player", parents=[root], location=None)
    sys.set_property("player", player)
    player.set_property("last_connect_time", time(), inherited=True)
    player.set_property("ownership_quota", 1000)
    player.set_property("gender", "neuter", inherited=True)
    player.set_property("ps", "it", inherited=True)
    player.set_property("po", "it", inherited=True)
    player.set_property("pp", "its", inherited=True)
    player.set_property("pr", "itself", inherited=True)
    player.set_property("pq", "its", inherited=True)
    player.set_property("psc", "It", inherited=True)
    player.set_property("poc", "It", inherited=True)
    player.set_property("ppc", "Its", inherited=True)
    player.set_property("prc", "Itself", inherited=True)
    player.set_property("pqc", "Its", inherited=True)
    player.set_property("home", None, inherited=True)
    player.set_property("gaglist", [], inherited=True)
    player.set_property("paranoid", False, inherited=True)
    player.set_property("responsible", [], inherited=True)
    player.set_property("lines", 10, inherited=True)
    player.set_property("object_gaglist", [], inherited=True)
    player.set_property("page_absent_msg", "%n is not currently logged in.", inherited=True)
    player.set_property("page_origin_msg", "You sense that %s is looking for you in %s.", inherited=True)
    player.set_property("page_echo_msg", "Your message has been sent.", inherited=True)
    player.set_property("whereis_location_msg", "%s is in %s.", inherited=True)
    player.set_property("who_location_msg", "%l", inherited=True)

    programmers = create("Generic Programmer", parents=[player], location=None)
    sys.set_property("programmer", programmers)
    wizards = create("Generic Wizard", parents=[programmers], location=None)
    sys.set_property("wizard", wizards)
    wizard.parents.add(wizards)

    thing = create("Generic Thing", parents=[root], location=None)
    sys.set_property("thing", thing)
    containers.parents.add(thing)

    exits = create("Generic Exit", parents=[root], location=None)
    sys.set_property("exit", exits)
    exits.set_property("open", False, inherited=True)
    exits.set_property("locked", False, inherited=True)
    exits.set_property("autolock", False, inherited=True)

    exits.set_property("source", None, inherited=True)
    exits.set_property("dest", None, inherited=True)
    exits.set_property("nogo_msg", "You can't go that way.", inherited=True)
    exits.set_property("onogo_msg", "{actor} can't go that way.", inherited=True)
    exits.set_property("arrive_msg", "You arrive at {subject}.", inherited=True)
    exits.set_property("oarrive_msg", "{actor} arrives at {subject}.", inherited=True)
    exits.set_property("oleave_msg", "{actor} leaves {subject}.", inherited=True)
    exits.set_property("leave_msg", "You leave {subject}.", inherited=True)

    notes = create("Generic Note", parents=[thing], location=None)
    sys.set_property("note", notes)
    letters = create("Generic Letter", parents=[notes], location=None)
    sys.set_property("letter", letters)

    lock_utils = create("Lock Utilities", parents=[notes], location=None)
    sys.set_property("lock_utils", lock_utils)

    lab = create("The Laboratory", parents=[rooms], location=None)
    lab.set_property(
        "description",
        """A cavernous laboratory filled with gadgetry of every kind,
    this seems like a dumping ground for every piece of dusty forgotten
    equipment a mad scientist might require.""",
    )

    wizard.location = lab
    wizard.save()

    new_player = create(name="Player", unique_name=True, location=lab)
    p = Player.objects.create()
    p.avatar = new_player
    p.save()

    new_player.parents.add(player, containers)
    new_player.owner = new_player
    new_player.save()

    root.get_verb("accept").delete()

    bootstrap.load_verbs(repo, "moo.core.bootstrap.default_verbs")
